/* 
*   lca: Linear Cellular Automata
*   88/04/25 john h. dubois iii (john@armory.com)
*   90/05/01 ported to XENIX
*   93/12/03 Added -h option
*/
/* See lca.doc for complete description */
#ifdef _M_SYSV
#define M_SYS5
#endif
#include <stdio.h>
#define ExTest(Test,Mess) if (Test) {fputs(Mess,stderr); exit(2);}
#define TRUE -1
#define FALSE 0
#ifdef __MSDOS__
#define FLAG_SPECS "-/" /* If in MSDOS, either '-' or '/' can specify a flag */
#else 
#define FLAG_SPECS "-"
#ifndef M_SYS5
#define strchr index	/* Use index function in cc */
#define rand random	/* use new random function */
#include <signal.h>	/* signalling facilities */
#endif
#endif
char *strchr();
char *UsageMess[] = {
"lca: display linear cellular automata.",
"Usage: lca [-h] [-r<rules>] [-c<char rules>] [-a<assoc chars>] [-w<width>]",
"           [-i<init string>] [-g<generations>] [-m<mutation period>]",
NULL};
char *Comment = "@(#) lca 1.1 96/07/03";

void CompressLine();

void main(argc,argv)
int argc;
char *argv[];
{
   /* Set chars associated with each state to default values */
   char *AssocChars = " .+X";
   int NumStates;	/* Number of possible states for each cell */
   int Width = 79;	/* Set number of cells to default value */
   int *Rules;		/* Rules used for determining next value of cells */
   int *Cells;		/* Array of integers representing cells */
   int Gens = 0;	/* Number of generations to print, set to default value
    which indicates that printing should continue until program is halted */
   int MutPeriod = 25;	/* Mutation period, set to default */
   long Time;		/* Used in seeding random number generator */
   int Beep = 0;	/* beep when cell line dies */

#ifdef __MSDOS__
   setvbuf(stdout, NULL, _IOLBF, 0x200);
   setcbrk(1);	/* Turn on control-break checking at all system calls */
   srand(time(&Time));	/* Seed random number generator */
#else 
#ifdef M_SYS5
   srand(time(&Time));	/* Seed random number generator */
#else
   signal(SIGINT,SIG_DFL); /* set interrupt signal to default */
   signal(SIGQUIT,SIG_DFL); /* set quit signal to default */
   srandom(getpid());	/* Use pid to seed rand # gen */
#endif
#endif
   /* Remove - or / from in front of command line arguments */
   RemoveFlagSpecs(argv + 1,argc - 1,FLAG_SPECS);
   /* Use flags to set up variables used to generate lca */
   if (!InterpretFlags(argv + 1,argc - 1,&AssocChars,&NumStates,&Width,&Cells,
    &Rules,&Gens,&MutPeriod,&Beep))
      /* If bad values given on command line, print usage */
      PrintStrings(UsageMess);
   else
      LCA(AssocChars,NumStates,Width,Cells,Rules,Gens,MutPeriod,Beep);
}

InterpretFlags(Args,Num,AssocChars,NumStates,Width,Cells,Rules,Gens,MutPeriod,
Beep)
/* Uses the command line arguments to set up variables used to generate lca */
char *Args[],	/* Command line args */
**AssocChars;	/* Chars associated with cell values, set to default string;
		   changed only if -a argument is given */
int Num,	/* Number of arguments given on command line */
*NumStates,	/* Number of possible states of cells, as determined by -a */
*Width,		/* Number of cells, already set to default; changed only if -w
		   argument is given */
**Rules,	/* Rules used to generate new values of cells */
*Gens,		/* Number of generations to print */
*MutPeriod,	/* Number of generations between mutations */
**Cells;	/* Integer representations of cells */
int *Beep;	/* whether to beep when cell line dies */
{
   int I;
   int c,i,r;	/* Whether -c, -i, and -r arguments were given */
   int Error,	/* Whether a bad value was given on command line */
   BadValArg;	/* Whether a bad value was given with last argument */
   char *RulesList;	/* String of integer or character rules */
   int NumRules;	/* Number of rules needed */
   char *InitStr;	/* Cell initialization string */

   Error = c = i = r = FALSE;	/* Start with flags false */
   for (I = 0; I < Num; I++) {	/* For each command line arg given */
      BadValArg = FALSE;
      switch (*Args[I]) {
	 case 'a':	*AssocChars = Args[I] + 1;
			BadValArg = strlen(*AssocChars) < 2;
			break;
	 case 'b':	*Beep = 1;
			break;
	 case 'c':	c = TRUE;
			RulesList = Args[I] + 1;
			break;
	 case 'g':	*Gens = atoi(Args[I] + 1);
			BadValArg = *Gens < 1;
			break;
	 case 'h':	Help();
			break;
	 case 'i':	i = TRUE;
			InitStr = Args[I] + 1;
			break;
	 case 'm':	*MutPeriod = atoi(Args[I] + 1);
			BadValArg = *MutPeriod < 0;
			break;
	 case 'r':	r = TRUE;
			RulesList = Args[I] + 1;
			break;
	 case 'w':	*Width = atoi(Args[I] + 1);
			BadValArg = *Width < 3;
			break;
	 default:	fputs("Bad command line argument: -",stdout);
			puts(Args[I]);
			Error = TRUE;
      }
      if (BadValArg) {
	 Error = TRUE;
	 fputs("Bad value given in -",stdout);
	 putchar(*Args[I]);
	 puts(" argument.");
      }
   }
   if (r && c) {
      puts("-r and -c arguments cannot be given together.");
      Error = TRUE;
   }
   if (Error)
      return(FALSE);
   *NumStates = strlen(*AssocChars);	/* Number of possible states of cells
    is equal to the number of associated chars given */
   NumRules = (*NumStates - 1) * 3 + 1;
   *Rules = (int *) calloc(NumRules,sizeof(int));
   if (r) {	/* Rules are given as integers */
      if (!IntRules(*Rules,RulesList,NumRules,*NumStates)) {
	 puts("Bad value given in -r argument.");
	 return(FALSE);
      }
   }
   else if (c) {	/* Rules are given as chars */
      if (!CharRules(*Rules,RulesList,NumRules,*AssocChars)) {
	 puts("Bad char given in -c argument.");
	 return(FALSE);
      }
   }
   else /* No rules given; initialize randomly */
      RandRules(*Rules,NumRules,*NumStates);
   *Cells = (int *) calloc(*Width,sizeof(int));
   if (i) {	/* Initialization string given; convert chars to their
		   integer cell value equivalents */
      if (!CharInit(*Cells,*Width,*AssocChars,InitStr)) {
	 puts("Bad char given in -i argument.");
	 return(FALSE);
      }
   }
   else /* No init string given; initialize cells randomly */
      RandInit(*Cells,*Width,*NumStates);
   return(TRUE);
}

CharInit(Cells,Width,AssocChars,InitStr)
/* Converts a string of char cell values to integer cell values by looking up
   chars in AssocChars */
int Cells[],	/* Cells to be set */
Width;		/* Number of cells */
char AssocChars[],	/* Chars associated with cell values */
InitStr[];		/* Initialization string */
{
   int I;
   int NumGiven;	/* Number of initialization chars given */
   char *MatchChar;	/* Pointer to matching char found in AssocChars */

   /* Find number of init chars given; do not allow it to be more than number
       of cells */
   NumGiven = Min(strlen(InitStr),Width);
   for (I = 0; I < NumGiven; I++) {
      MatchChar = strchr(AssocChars,InitStr[I]);	/* Look up char */
      if (MatchChar == NULL)	/* If invalid char, return with error */
	 return(FALSE);
      else	/* Find value of cell (its offset in the array) */
	 Cells[I] = MatchChar - AssocChars;
   }
   for (; I < Width; I++)  /* If not enough chars given, set the rest to 0 */
      Cells[I] = 0;
   return(TRUE);
}

RandInit(Cells,Width,NumStates)
/* Initialize cells with random values */
int Cells[],	/* Cells to be initialized */
Width,		/* Number of cells */
NumStates;	/* Number of states that a cell can have */
{
   int I;

   for (I = 0; I < Width; I++)
      Cells[I] = rand() % NumStates;
}

IntRules(Rules,RulesList,NumRules,NumStates)
/* Convert a string of char representations of integer values separated by
   commas to numeric integers and use them to initalize lca rules */
int *Rules,	/* lca rules */
NumRules,	/* Number of rules needed */
NumStates;	/* Number of possible states of cells */
char *RulesList;	/* String of integers */
{
   int I;
   char *Rule;	/* Current string being converted to int */
   char *strtok();	/* String tokenizing function */

   Rule = strtok(RulesList,",");	/* Start tokenization */
   for (I = 0; I < NumRules && Rule != NULL; I++) {
      Rules[I] = atoi(Rule);	/* Convert string to int */
      if (Rules[I] < 0 || Rules[I] >= NumStates)
	 return(FALSE);
      Rule = strtok(NULL,",");	/* Get next int substring */
   }
   for (; I < NumRules; I++) /* If not enough rules given, set the rest to 0 */
      Rules[I] = 0;
   return(TRUE);
}

CharRules(Rules,RulesList,NumRules,AssocChars)
/* Converts a string of chars representing cell values to their integer
   equivalents by looking them up in a list of associtated chars and
   intializes lca rules with them */
int *Rules,	/* lca rules */
NumRules;	/* Number of rules needed */
char AssocChars[],	/* Chars associated with each cell value */
RulesList[];	/* Chars representing rules */
{
   int I;
   int NumGiven;	/* Number of char rules given */
   char *MatchChar;	/* Pointer to matching char found */

   /* Find number of rules given; do not allow it to be larger than the
      number of rules needed */
   NumGiven = Min(strlen(RulesList),NumRules);
   for (I = 0; I < NumGiven; I++) {
      MatchChar = strchr(AssocChars,RulesList[I]);
      if (MatchChar == NULL)
      /* If char not found in lookup, return with error */
	 return(FALSE);
      else	/* Find value (offset of char in lookup string) */
	 Rules[I] = MatchChar - AssocChars;
   }
   for (; I < NumRules; I++)
      /* If not enough rules given, initialize the rest to 0 */
      Rules[I] = 0;
   return(TRUE);
}

Min(I1,I2)
/* Returns the minimum of two integers */
int I1,I2;
{
   if (I1 < I2)
      return(I1);
   else
      return(I2);
}

RandRules(Rules,NumRules,NumStates)
/* Initializes rules array with random values */
int Rules[],	/* Rules array */
NumRules,	/* Number of rules needed */
NumStates;	/* Number of possible states of a cell */
{
   int I;

   for (I = 0; I < NumRules; I++)
      Rules[I] = rand() % NumStates;
}

RemoveFlagSpecs(Args,Num,FlagSpecs)
/* Removes any leading '-' or '/' from command line arguments by incrementing
   the argument pointer */
char *Args[];
int Num;
char FlagSpecs[];
{
   for (; Num; Num--)
      if (strchr(FlagSpecs,*(Args[Num - 1])) != NULL)
	 Args[Num - 1]++;
}

PrintStrings(Strings)
/* Prints an array of strings, one per line, until a terminating NULL pointer
   is found */
char *Strings[];
{
   int I;

   for (I = 0; Strings[I] != NULL; I++)
      puts(Strings[I]);
}

LCA(AssocChars,NumStates,Width,Cells,Rules,Gens,MutPeriod,Beep)
/* Generates and prints generations of linear cellular automata */
char AssocChars[];	/* Chars associated with each cell value */
int NumStates,		/* Number of possible states for a cell */
Width,		/* Number of cells */
Cells[],	/* Cell array */
Rules[],	/* Rules for generating succeeding generations of cells */
Gens,		/* Number of generations to produce */
MutPeriod;	/* Mutation period */
int Beep;	/* beep when cells die */
{
   /* If Gens is 0, the program should continue until halted */
   int NoLimit = !Gens;
   int MutTimer = MutPeriod;	/* Countdown for mutation */
   int NumRules = (NumStates - 1) * 3 + 1;	/* Number of rules needed */
   char *CharCells;	/* Character representation of a generation of cells */
   int *Middle; /* Pointer to the middle of the cells array */

   Middle = Cells + (Width >> 1);	/* Find middle of cells array */
   CharCells = (char *) calloc(Width + 1,sizeof(char));
   while (NoLimit || Gens--) {
      /* Convert cells to char string */
      ConvertCells(Cells,AssocChars,Width,CharCells);
      /* Compress spaces in line to tabs; delete trailing spaces and tabs */
      CompressLine(CharCells);
      puts(CharCells);	/* Print char representation of cells */
      NextGen(Cells,Rules,Width);	/* Find next generation of cells */
      if (MutPeriod && !(--MutTimer)) { /* If time to do mutation */
	 MutTimer = MutPeriod;	/* Reset mutation countdown */
	 RandRules(Rules,NumRules,NumStates);	/* Get new random rules */
	 /* If all cells are identical, change the middle two cells */
	 if (AllSame(Cells,Width)) {
	    *Middle = NumStates - 1;
	    *(Middle + 1) = NumStates - 2;
	    if (Beep)
		putchar('\07');	/* Beep */
	 }
      }
   }
}

AllSame(Cells,Width)
/* Returns true if all of the cells in Cells are identical */
int Cells[],Width;
{
   int I;
   int RightMost = Cells[Width - 1];

   for (I = Width - 2; I; I--)
      if (Cells[I] != RightMost)
	 return(FALSE);
   return(*Cells == RightMost);
}

NextGen(Cells,Rules,Width)
/* Generates the next generation of cells using Rules */
int Cells[],	/* Cells array */
Rules[],	/* Rules for generating new cells */
Width;		/* Number of cells */
{
   int I;
   int TempL,TempM;	/* Temporary storage of cells to the left of and the
    right of the current three cells being examined and used */
   int Limit = Width - 1;	/* Highest cell to modify in loop */
   int LeftMost;	/* Original value of leftmost cell */

   /* Start with the rightmost cell value as TempL because the cells are in a
      circular configuration and the leftmost cell will be the first
      calculated; the rightmost cell is the cell to the left of the leftmost
      cell */
   TempL = Cells[Width - 1];
   /* Save value of leftmost cell for calculating value of rightmost cell */
   LeftMost = Cells[0];
   for (I = 0; I < Limit; I++) {
      TempM = Cells[I];
      Cells[I] = Rules[TempL + TempM + Cells[I + 1]];
      TempL = TempM;
   }
   /* Determine value of rightmost cell */
   Cells[Limit] = Rules[TempL + Cells[Limit] + LeftMost];
}

ConvertCells(Cells,AssocChars,Width,CharCells)
/* Convert the integer values of cells to characters that can be printed */
int *Cells,	/* Integer cell values */
Width;		/* Number of cells */
char AssocChars[],	/* Chars associated with each possible cell value */
CharCells[];		/* Array to put chars in */
{
   int I;

   for (I = 0; I < Width; I++)
      CharCells[I] = AssocChars[Cells[I]];
   CharCells[Width] = '\0';
}

void CompressLine(Line)
/* Replaces spaces in Line with a tab wherever possible (except replacement
   is not done if a single space would be replaced by a tab); deletes
   trailing spaces and tabs.  All control chars in Line except tabs are taken
   to occupy one space. */
char *Line;
{
   int CharCount = 0;	/* How many chars have been found since the last
			   tab position */
   int I;		/* Source char being converted */
   int NewPos = -1;	/* Position in line that chars are being written to */

   for (I = 0; Line[I] != '\0'; I++) {
      CharCount++;
      if (Line[I] == '\t')	/* If char is a tab, delete any immediately
	       preceding spaces within the tab block that it will skip */
	 while (--CharCount && Line[NewPos] == ' ')
	    NewPos--;
      Line[++NewPos] = Line[I]; /* Copy char to new position */
      if (CharCount == 8) {	/* If end of a tab block has been reached */
	 if (Line[NewPos] == ' ' && Line[NewPos - 1] == ' ') {
	    /* If at least 2 spaces would be replaced, delete trailing
	       spaces in the tab block and write a tab instead */
	    NewPos--;
	    while (--CharCount && Line[--NewPos] == ' ')
	       ;
	    Line[++NewPos] = '\t';
	 }
	 CharCount = 0; /* Reset char count */
      }
   }
   while ((Line[NewPos] == ' ' || Line[NewPos] == '\t') && NewPos >= 0)
      /* Delete trailing spaces and tabs */
      NewPos--;
   Line[++NewPos] = '\0';	/* Add terminator */
}

char *strtok(S,Sep)
char *S;
char *Sep;
/* S is a sequence of zero or more text tokens, separated by spans of one or
   more characters from the separator string Sep.
   The first call to strtok returns a pointer to the first character of the
   first token in S and writes a null char into S immediately following the
   returned token.  Subsequent calls with NULL for the first argument will
   work through the string in this way until no tokens remain.
   The separator string may be different from call to call.
   When no tokens remain in S, a NULL pointer is returned. */
{
   static char *Toks;
   char *Temp;

   if (S != NULL)	
      /* If new string is being tokenized, set the static pointer to it */
      Toks = S;
   while (*Toks != '\0' && strchr(Sep,*Toks) != NULL)
      Toks++;	/* Get rid of leading separator chars */
   if (*Toks == '\0')	/* If no more tokens, return NULL */
      return(NULL);
   /* Otherwise set pointer that will be returned to start of token */
   Temp = Toks;		
   while (*Toks != '\0' && strchr(Sep,*Toks) == NULL)
      Toks++;	/* Find end of token */
   if (*Toks != '\0')	/* If token did not end with null char, terminate it
   			   with a null char and increment tokenizing pointer */
      *(Toks++) = '\0';
   return(Temp);
}

Help()
{
    char *HelpMsg[] = {
"   lca: generate linear cellular automata.",
"   Usage: lca [options]",
"   This program calculates and displays succeeding generations of linear",
"cellular automata.  LCA are related to the Game of Life, but are one- rather",
"than two-dimensional.  Also, while in the Game of Life each cell is in one",
"of two states (alive or dead), LCA may have many states, and in this",
"implementation the number of states is not fixed.",
"   The state of each cell is refered to as its value.  Each cell has a value",
"between zero and one less than the number of possible states.  To find the",
"value of the cell in the next generation, its current value and the current",
"values of the cells on either side are added together.  The result is used",
"to select a \"rule\".  There is one rule for each possible result.  Thus, if",
"each cell has four possible values (0-3), the sum of the three cells will be",
"between 0 and 9, and there will be ten rules.  Each rule is a value which",
"will be the new value of the cell.  For example, if the values of a cell,",
"its left neighbor, and it right neighbor are 2, 0, and 3, and the value of",
"rule 5 is 1, then the value of the cell in the next generation will be 1.",
"The cells are arranged in a circular configuration; that is, the leftmost",
"and rightmost cells are taken to be neighbors.",
"   To display the cells, a unique character is associated with each value,",
"and the character associated with a particular value is what is printed to",
"represent a cell of that value.  Thus, the cells are displayed in the form",
"of a string of characters.  The number of cells is typically chosen to be",
"one less than the width of the display device that will be used, for",
"convenience in display and to avoid problems with line wrap.  Succeeding",
"generations of cells will be printed on succeeding lines.  The two-",
"dimentional, ongoing pattern produced by succeeding generations of cells can",
"be extremely interesting.",
"   The following arguments are used to determine the nature of the LCA:",
"   -a<associated chars> : a is followed by a string of chars.  The leftmost",
"char should be the character to be associated with value (state) 0; the next",
"char should be the character to be associated with value 1, etc.  The total",
"number of chars will determine the number of possible states.  If -a is not",
"given, the associated char string defaults to \" .+X\".",
"   -r<rules> : r is followed by a series of integers separated by commas.",
"The leftmost number should be the value of rule 0; the next, the value of",
"rule 1, etc.  The number of rules given should be (1 + 3 * [highest possible",
"value of a cell, as determined by the -a argument]), since there must be one",
"rule for each possible value of the sum of the three cells.  For example, if",
"four chars are given with the -a argument, then the possible values of a",
"cell are 0 - 3, and there should be 10 rules given with the -r argument (one",
"for each of the possible sums 0 - 9).  If there are not enough rules, the",
"highest rules will be set to 0.",
"   -c<char rules> : The rules may be specified by giving string of",
"characters instead of integers.  Each character will be looked up in the",
"string given in the -a argument to find its value.",
"   If neither -r nor -c is given, the rules are initialized to random values.",
"   -w<width> : width should be an integer specifying the number of cells.",
"The default is 79.",
"   -i<init string> : The initial values of the cells can be given as a",
"string of characters which are translated to values in the same manner as",
"the characters given in the -c argument are.  If not enough characters are",
"given, the rightmost cells will be set to 0.  If this argument is not given,",
"the cells are initialized randomly.",
"   -g<generations> : The number of generations to produce (and thus lines to",
"print).  If this argument is not given, the program will continue until halted.",
"   -m<mutation period> : <mutation period> should be an integer.  The rules",
"will be replaced with a new random set each time that the specified period",
"(number of generations) has elapsed.  The default is 25.  If 0 is specified,",
"no mutations will occur.  Each time a mutation occurs, the cells are checked",
"to determine whether they are all identical.  If they are, the value of the",
"middle cell is changed to <maxcellvalue> and the value of the cell to the",
"right of the middle cell is changed to <maxcellvalue> - 1, in order to begin",
"an interesting sequence again.  The terminal will beep when this occurs.",
"   Author: John DuBois (john@armory.com)",
NULL};

    PrintStrings(HelpMsg);
    exit(0);
}
